@@ -1,6 +1,6 @@
 /*
  *  Phusion Passenger - https://www.phusionpassenger.com/
- *  Copyright (c) 2010, 2011, 2012 Phusion
+ *  Copyright (c) 2010-2013 Phusion
  *
  *  "Phusion Passenger" is a trademark of Hongli Lai & Ninh Bui.
  *
@@ -40,6 +40,7 @@
 #include <dirent.h>
 #include <limits.h>
 #include <unistd.h>
+#include <string.h>
 #include <signal.h>
 #ifdef __linux__
 	#include <sys/syscall.h>
@@ -400,6 +401,34 @@ getProcessUsername() {
 	}
 }
 
+string
+getGroupName(gid_t gid) {
+	struct group *groupEntry;
+
+	groupEntry = getgrgid(gid);
+	if (groupEntry == NULL) {
+		return toString(gid);
+	} else {
+		return groupEntry->gr_name;
+	}
+}
+
+gid_t
+lookupGid(const string &groupName) {
+	struct group *groupEntry;
+
+	groupEntry = getgrnam(groupName.c_str());
+	if (groupEntry == NULL) {
+		if (looksLikePositiveNumber(groupName)) {
+			return atoi(groupName);
+		} else {
+			return (gid_t) -1;
+		}
+	} else {
+		return groupEntry->gr_gid;
+	}
+}
+
 mode_t
 parseModeString(const StaticString &mode) {
 	mode_t modeBits = 0;
@@ -794,7 +823,7 @@ prestartWebApps(const ResourceLocator &locator, const string &ruby,
 }
 
 void
-runAndPrintExceptions(const function<void ()> &func, bool toAbort) {
+runAndPrintExceptions(const boost::function<void ()> &func, bool toAbort) {
 	try {
 		func();
 	} catch (const boost::thread_interrupted &) {
@@ -808,7 +837,7 @@ runAndPrintExceptions(const function<void ()> &func, bool toAbort) {
 }
 
 void
-runAndPrintExceptions(const function<void ()> &func) {
+runAndPrintExceptions(const boost::function<void ()> &func) {
 	runAndPrintExceptions(func, true);
 }
 
@@ -819,10 +848,11 @@ getHostName() {
 		// https://bugzilla.redhat.com/show_bug.cgi?id=130733
 		hostNameMax = 255;
 	}
-	char hostname[hostNameMax + 1];
-	if (gethostname(hostname, hostNameMax) == 0) {
-		hostname[hostNameMax] = '\0';
-		return hostname;
+
+	string buf(hostNameMax + 1, '\0');
+	if (gethostname(&buf[0], hostNameMax + 1) == 0) {
+		buf[hostNameMax] = '\0';
+		return string(buf.c_str());
 	} else {
 		int e = errno;
 		throw SystemException("Unable to query the system's host name", e);
