@@ -30,6 +30,8 @@
 #include <boost/foreach.hpp>
 #include <boost/enable_shared_from_this.hpp>
 #include <string>
+#include <utility>
+#include <vector>
 
 #include <sys/select.h>
 #include <sys/types.h>
@@ -54,6 +56,7 @@
 #include <Utils/Base64.h>
 #include <Utils/Timer.h>
 #include <Utils/ScopeGuard.h>
+#include <Utils/StrIntUtils.h>
 #include <Utils/IOUtils.h>
 #include <Utils/MessageIO.h>
 #include <Utils/VariantMap.h>
@@ -72,11 +75,13 @@ enum OomFileType {
 #define REQUEST_SOCKET_PASSWORD_SIZE     64
 
 class ServerInstanceDirToucher;
+class AgentWatcher;
 static bool hasEnvOption(const char *name, bool defaultValue = false);
 static void setOomScore(const StaticString &score);
 
 
-/** The options that were passed to AgentsStarter. */
+/***** Agent options *****/
+
 static VariantMap agentsOptions;
 static string  tempDir;
 static bool    userSwitching;
@@ -84,31 +89,37 @@ static string  defaultUser;
 static string  defaultGroup;
 static uid_t   webServerWorkerUid;
 static gid_t   webServerWorkerGid;
-static string  serializedPrestartURLs;
 
 /***** Working objects *****/
+
+struct WorkingObjects {
+	RandomGenerator randomGenerator;
+	EventFd errorEvent;
+	ResourceLocatorPtr resourceLocator;
+	ServerInstanceDirPtr serverInstanceDir;
+	ServerInstanceDir::GenerationPtr generation;
+	uid_t defaultUid;
+	gid_t defaultGid;
+	vector<string> cleanupPidfiles;
+	string loggingAgentAddress;
+	string loggingAgentPassword;
+	string loggingAgentAdminAddress;
+	string adminToolStatusPassword;
+	string adminToolManipulationPassword;
+};
+
+typedef boost::shared_ptr<WorkingObjects> WorkingObjectsPtr;
+
 static string oldOomScore;
-static RandomGenerator *randomGenerator;
-static EventFd *errorEvent;
-static ResourceLocator *resourceLocator;
-static ServerInstanceDirPtr serverInstanceDir;
-static ServerInstanceDir::GenerationPtr generation;
-static ServerInstanceDirToucher *serverInstanceDirToucher;
-static uid_t defaultUid;
-static gid_t defaultGid;
-static string loggingAgentAddress;
-static string loggingAgentPassword;
-static string loggingAgentAdminAddress;
-static string adminToolStatusPassword;
-static string adminToolManipulationPassword;
 
 #include "AgentWatcher.cpp"
+#include "ServerInstanceDirToucher.cpp"
 #include "HelperAgentWatcher.cpp"
 #include "LoggingAgentWatcher.cpp"
-#include "ServerInstanceDirToucher.cpp"
-static vector<AgentWatcherPtr> watchers;
 
 
+/***** Functions *****/
+
 static bool
 hasEnvOption(const char *name, bool defaultValue) {
 	const char *value = getenv(name);
@@ -218,19 +229,19 @@ setOomScoreNeverKill() {
  * an error.
  */
 static bool
-waitForStarterProcessOrWatchers(vector<AgentWatcherPtr> &watchers) {
+waitForStarterProcessOrWatchers(const WorkingObjectsPtr &wo, vector<AgentWatcherPtr> &watchers) {
 	fd_set fds;
 	int max, ret;
 	char x;
 	
 	FD_ZERO(&fds);
 	FD_SET(FEEDBACK_FD, &fds);
-	FD_SET(errorEvent->fd(), &fds);
+	FD_SET(wo->errorEvent.fd(), &fds);
 	
-	if (FEEDBACK_FD > errorEvent->fd()) {
+	if (FEEDBACK_FD > wo->errorEvent.fd()) {
 		max = FEEDBACK_FD;
 	} else {
-		max = errorEvent->fd();
+		max = wo->errorEvent.fd();
 	}
 	
 	ret = syscalls::select(max + 1, &fds, NULL, NULL, NULL);
@@ -240,7 +251,7 @@ waitForStarterProcessOrWatchers(vector<AgentWatcherPtr> &watchers) {
 		return false;
 	}
 	
-	if (FD_ISSET(errorEvent->fd(), &fds)) {
+	if (FD_ISSET(wo->errorEvent.fd(), &fds)) {
 		vector<AgentWatcherPtr>::const_iterator it;
 		string message, backtrace, watcherName;
 		
@@ -263,83 +274,125 @@ waitForStarterProcessOrWatchers(vector<AgentWatcherPtr> &watchers) {
 	}
 }
 
+static vector<pid_t>
+readCleanupPids(const WorkingObjectsPtr &wo) {
+	vector<pid_t> result;
+
+	foreach (string filename, wo->cleanupPidfiles) {
+		FILE *f = fopen(filename.c_str(), "r");
+		if (f != NULL) {
+			char buf[33];
+			size_t ret;
+
+			ret = fread(buf, 1, 32, f);
+			if (ret > 0) {
+				buf[ret] = '\0';
+				result.push_back(atoi(buf));
+			} else {
+				P_WARN("Cannot read cleanup PID file " << filename);
+			}
+		} else {
+			P_WARN("Cannot open cleanup PID file " << filename);
+		}
+	}
+
+	return result;
+}
+
 static void
-cleanupAgentsInBackground(vector<AgentWatcherPtr> &watchers, char *argv[]) {
+cleanupAgentsInBackground(const WorkingObjectsPtr &wo, vector<AgentWatcherPtr> &watchers, char *argv[]) {
 	this_thread::disable_interruption di;
 	this_thread::disable_syscall_interruption dsi;
+	vector<pid_t> cleanupPids;
 	pid_t pid;
 	int e;
 
+	cleanupPids = readCleanupPids(wo);
+
 	pid = fork();
 	if (pid == 0) {
 		// Child
-		vector<AgentWatcherPtr>::const_iterator it;
-		Timer timer(false);
-		fd_set fds, fds2;
-		int max, agentProcessesDone;
-		unsigned long long deadline = 30000; // miliseconds
+		try {
+			vector<AgentWatcherPtr>::const_iterator it;
+			Timer timer(false);
+			fd_set fds, fds2;
+			int max, agentProcessesDone;
+			unsigned long long deadline = 30000; // miliseconds
 
-		#if defined(__linux__) || defined(__APPLE__) || defined(__FreeBSD__) || defined(sun)
-			// Change process title.
-			strcpy(argv[0], "PassengerWatchdog (cleaning up...)");
-		#endif
-		
-		// Wait until all agent processes have exited. The starter
-		// process is responsible for telling the individual agents
-		// to exit.
-		
-		max = 0;
-		FD_ZERO(&fds);
-		for (it = watchers.begin(); it != watchers.end(); it++) {
-			FD_SET((*it)->getFeedbackFd(), &fds);
-			if ((*it)->getFeedbackFd() > max) {
-				max = (*it)->getFeedbackFd();
-			}
-		}
-		
-		timer.start();
-		agentProcessesDone = 0;
-		while (agentProcessesDone != -1
-		    && agentProcessesDone < (int) watchers.size()
-		    && timer.elapsed() < deadline)
-		{
-			struct timeval timeout;
-			
-			#ifdef FD_COPY
-				FD_COPY(&fds, &fds2);
-			#else
-				FD_ZERO(&fds2);
-				for (it = watchers.begin(); it != watchers.end(); it++) {
-					FD_SET((*it)->getFeedbackFd(), &fds2);
-				}
+			#if defined(__linux__) || defined(__APPLE__) || defined(__FreeBSD__) || defined(sun)
+				// Change process title.
+				strcpy(argv[0], "PassengerWatchdog (cleaning up...)");
 			#endif
 			
-			timeout.tv_sec = 0;
-			timeout.tv_usec = 10000;
-			agentProcessesDone = syscalls::select(max + 1, &fds2, NULL, NULL, &timeout);
-			if (agentProcessesDone > 0 && timer.elapsed() < deadline) {
-				usleep(10000);
+			// Wait until all agent processes have exited. The starter
+			// process is responsible for telling the individual agents
+			// to exit.
+			
+			max = 0;
+			FD_ZERO(&fds);
+			for (it = watchers.begin(); it != watchers.end(); it++) {
+				FD_SET((*it)->getFeedbackFd(), &fds);
+				if ((*it)->getFeedbackFd() > max) {
+					max = (*it)->getFeedbackFd();
+				}
 			}
+
+			timer.start();
+			agentProcessesDone = 0;
+			while (agentProcessesDone != -1
+			    && agentProcessesDone < (int) watchers.size()
+			    && timer.elapsed() < deadline)
+			{
+				struct timeval timeout;
+				
+				#ifdef FD_COPY
+					FD_COPY(&fds, &fds2);
+				#else
+					FD_ZERO(&fds2);
+					for (it = watchers.begin(); it != watchers.end(); it++) {
+						FD_SET((*it)->getFeedbackFd(), &fds2);
+					}
+				#endif
+				
+				timeout.tv_sec = 0;
+				timeout.tv_usec = 10000;
+				agentProcessesDone = syscalls::select(max + 1, &fds2, NULL, NULL, &timeout);
+				if (agentProcessesDone > 0 && timer.elapsed() < deadline) {
+					usleep(10000);
+				}
+			}
+
+			if (agentProcessesDone == -1 || timer.elapsed() >= deadline) {
+				// An error occurred or we've waited long enough. Kill all the
+				// processes.
+				P_WARN("Some Phusion Passenger agent processes did not exit " <<
+					"in time, forcefully shutting down all.");
+			} else {
+				P_DEBUG("All Phusion Passenger agent processes have exited. Forcing all subprocesses to shut down.");
+			}
+			for (it = watchers.begin(); it != watchers.end(); it++) {
+				(*it)->forceShutdown();
+			}
+
+			// Now clean up the server instance directory.
+			wo->generation->destroy();
+			wo->serverInstanceDir->destroy();
+
+			// Notify given PIDs about our shutdown.
+			foreach (pid_t pid, cleanupPids) {
+				P_DEBUG("Sending SIGTERM to cleanup PID " << pid);
+				kill(pid, SIGTERM);
+			}
+			strcpy(argv[0], "PassengerWatchdog (cleaning up 6...)");
+			_exit(0);
+		} catch (const std::exception &e) {
+			P_CRITICAL("An exception occurred during cleaning up: " << e.what());
+			_exit(1);
+		} catch (...) {
+			P_CRITICAL("An unknown exception occurred during cleaning up");
+			_exit(1);
 		}
 		
-		if (agentProcessesDone == -1 || timer.elapsed() >= deadline) {
-			// An error occurred or we've waited long enough. Kill all the
-			// processes.
-			P_WARN("Some Phusion Passenger agent processes did not exit " <<
-				"in time, forcefully shutting down all.");
-		} else {
-			P_DEBUG("All Phusion Passenger agent processes have exited. Forcing all subprocesses to shut down.");
-		}
-		for (it = watchers.begin(); it != watchers.end(); it++) {
-			(*it)->forceShutdown();
-		}
-		
-		// Now clean up the server instance directory.
-		delete generation.get();
-		delete serverInstanceDir.get();
-		
-		_exit(0);
-		
 	} else if (pid == -1) {
 		// Error
 		e = errno;
@@ -349,8 +402,8 @@ cleanupAgentsInBackground(vector<AgentWatcherPtr> &watchers, char *argv[]) {
 		// Parent
 		
 		// Let child process handle cleanup.
-		serverInstanceDir->detach();
-		generation->detach();
+		wo->serverInstanceDir->detach();
+		wo->generation->detach();
 	}
 }
 
@@ -371,17 +424,86 @@ inferDefaultGroup(const string &defaultUser) {
 			string("The user that PassengerDefaultUser refers to, '") +
 			defaultUser + "', does not exist.");
 	}
-	
-	struct group *groupEntry = getgrgid(userEntry->pw_gid);
-	if (groupEntry == NULL) {
-		throw ConfigurationException(
-			string("The option PassengerDefaultUser is set to '" +
-			defaultUser + "', but its primary group doesn't exist. "
-			"In other words, your system's user account database "
-			"is broken. Please fix it."));
+	return getGroupName(userEntry->pw_gid);
+}
+
+static vector< pair<string, string> >
+agentsOptionsToEnvVars(const VariantMap &agentsOptions) {
+	vector< pair<string, string> > result;
+	VariantMap::ConstIterator it, end = agentsOptions.end();
+
+	result.reserve(agentsOptions.size());
+	for (it = agentsOptions.begin(); it != end; it++) {
+		result.push_back(make_pair("passenger_" + it->first, it->second));
+	}
+
+	return result;
+}
+
+static void
+setEnvVarsFromVector(const vector< pair<string, string> > &envvars) {
+	vector< pair<string, string> >::const_iterator it;
+
+	for (it = envvars.begin(); it != envvars.end(); it++) {
+		setenv(it->first.c_str(), it->second.c_str(), 1);
+	}
+}
+
+static bool
+runHookScript(const char *name) {
+	TRACE_POINT();
+	string value = agentsOptions.get(string("hook_") + name, false);
+	if (value.empty()) {
+		return true;
+	}
+
+	vector< pair<string, string> > envvars = agentsOptionsToEnvVars(agentsOptions);
+	pid_t pid;
+	int e, status;
+
+	P_INFO("Running " << name << " hook script: " << value);
+
+	pid = fork();
+	if (pid == 0) {
+		resetSignalHandlersAndMask();
+		disableMallocDebugging();
+		closeAllFileDescriptors(2);
+		setEnvVarsFromVector(envvars);
+
+		execlp(value.c_str(), value.c_str(), (const char * const) 0);
+		e = errno;
+		fprintf(stderr, "*** ERROR: Cannot execute %s hook script %s: %s (errno=%d)\n",
+			name, value.c_str(), strerror(e), e);
+		fflush(stderr);
+		_exit(1);
+		return true; // Never reached.
+
+	} else if (pid == -1) {
+		e = errno;
+		P_ERROR("Cannot fork a process for hook script " << value <<
+			": " << strerror(e) << " (errno=" << e << ")");
+		return false;
+
+	} else if (waitpid(pid, &status, 0) == -1) {
+		e = errno;
+		P_ERROR("Unable to wait for hook script " << value <<
+			" (PID " << pid << "): " << strerror(e) << " (errno=" <<
+			e << ")");
+		return false;
+
+	} else {
+		P_INFO("Hook script " << value << " (PID " << pid <<
+			") exited with status " << WEXITSTATUS(status));
+		return WEXITSTATUS(status) == 0;
+	}
+}
+
+static void
+runHookScriptAndThrowOnError(const char *name) {
+	TRACE_POINT();
+	if (!runHookScript(name)) {
+		throw RuntimeException(string("Hook script ") + name + " failed");
 	}
-	
-	return groupEntry->gr_name;
 }
 
 static void
@@ -392,6 +514,7 @@ initializeBareEssentials(int argc, char *argv[]) {
 	 * forcefully redirect stdout to stderr so that everything ends up in the
 	 * same place.
 	 */
+	int oldStdout = dup(1);
 	dup2(2, 1);
 
 	/*
@@ -406,6 +529,12 @@ initializeBareEssentials(int argc, char *argv[]) {
 	oldOomScore = setOomScoreNeverKill();
 	
 	agentsOptions = initializeAgent(argc, argv, "PassengerWatchdog");
+
+	if (agentsOptions.get("test_binary", false) == "1") {
+		write(oldStdout, "PASS\n", 5);
+		exit(0);
+	}
+	close(oldStdout);
 }
 
 static void
@@ -463,7 +592,6 @@ maybeSetsid() {
 static void
 lookupDefaultUidGid(uid_t &uid, gid_t &gid) {
 	struct passwd *userEntry;
-	struct group  *groupEntry;
 	
 	userEntry = getpwnam(defaultUser.c_str());
 	if (userEntry == NULL) {
@@ -472,23 +600,21 @@ lookupDefaultUidGid(uid_t &uid, gid_t &gid) {
 	}
 	uid = userEntry->pw_uid;
 
-	groupEntry = getgrnam(defaultGroup.c_str());
-	if (groupEntry == NULL) {
+	gid = lookupGid(defaultGroup);
+	if (gid == (gid_t) -1) {
 		throw NonExistentGroupException("Default group '" + defaultGroup +
 			"' does not exist.");
 	}
-	gid = groupEntry->gr_gid;
 }
 
 static void
-initializeWorkingObjects() {
+initializeWorkingObjects(WorkingObjectsPtr &wo, ServerInstanceDirToucherPtr &serverInstanceDirToucher) {
 	TRACE_POINT();
-	randomGenerator = new RandomGenerator();
-	errorEvent = new EventFd();
-	resourceLocator = new ResourceLocator(agentsOptions.get("passenger_root"));
+	wo = boost::make_shared<WorkingObjects>();
+	wo->resourceLocator = boost::make_shared<ResourceLocator>(agentsOptions.get("passenger_root"));
 
 	UPDATE_TRACE_POINT();
-	// Must not used make_shared() here because Watchdog.cpp
+	// Must not used boost::make_shared() here because Watchdog.cpp
 	// deletes the raw pointer in cleanupAgentsInBackground().
 	if (agentsOptions.get("server_instance_dir", false).empty()) {
 		/* We embed the super structure version in the server instance directory name
@@ -501,54 +627,58 @@ initializeWorkingObjects() {
 			toString(SERVER_INSTANCE_DIR_STRUCTURE_MAJOR_VERSION) + "." +
 			toString(SERVER_INSTANCE_DIR_STRUCTURE_MINOR_VERSION) + "." +
 			toString<unsigned long long>(agentsOptions.getPid("web_server_pid"));
-		serverInstanceDir.reset(new ServerInstanceDir(path));
+		wo->serverInstanceDir.reset(new ServerInstanceDir(path));
 	} else {
-		serverInstanceDir.reset(new ServerInstanceDir(agentsOptions.get("server_instance_dir")));
-		agentsOptions.set("server_instance_dir", serverInstanceDir->getPath());
+		wo->serverInstanceDir.reset(new ServerInstanceDir(agentsOptions.get("server_instance_dir")));
+		agentsOptions.set("server_instance_dir", wo->serverInstanceDir->getPath());
 	}
-	generation = serverInstanceDir->newGeneration(userSwitching, defaultUser,
+	wo->generation = wo->serverInstanceDir->newGeneration(userSwitching, defaultUser,
 		defaultGroup, webServerWorkerUid, webServerWorkerGid);
-	agentsOptions.set("server_instance_dir", serverInstanceDir->getPath());
-	agentsOptions.setInt("generation_number", generation->getNumber());
+	agentsOptions.set("server_instance_dir", wo->serverInstanceDir->getPath());
+	agentsOptions.setInt("generation_number", wo->generation->getNumber());
+	agentsOptions.set("generation_path", wo->generation->getPath());
 
 	UPDATE_TRACE_POINT();
-	serverInstanceDirToucher = new ServerInstanceDirToucher();
+	serverInstanceDirToucher = boost::make_shared<ServerInstanceDirToucher>(wo);
 
 	UPDATE_TRACE_POINT();
-	lookupDefaultUidGid(defaultUid, defaultGid);
+	lookupDefaultUidGid(wo->defaultUid, wo->defaultGid);
 
 	UPDATE_TRACE_POINT();
-	loggingAgentAddress  = "unix:" + generation->getPath() + "/logging";
-	loggingAgentPassword = randomGenerator->generateAsciiString(64);
-	loggingAgentAdminAddress  = "unix:" + generation->getPath() + "/logging_admin";
+	wo->cleanupPidfiles = agentsOptions.getStrSet("cleanup_pidfiles", false);
 
 	UPDATE_TRACE_POINT();
-	adminToolStatusPassword = randomGenerator->generateAsciiString(MESSAGE_SERVER_MAX_PASSWORD_SIZE);
-	adminToolManipulationPassword = randomGenerator->generateAsciiString(MESSAGE_SERVER_MAX_PASSWORD_SIZE);
-	agentsOptions.set("admin_tool_status_password", adminToolStatusPassword);
-	agentsOptions.set("admin_tool_manipulation_password", adminToolManipulationPassword);
+	wo->loggingAgentAddress  = "unix:" + wo->generation->getPath() + "/logging";
+	wo->loggingAgentPassword = wo->randomGenerator.generateAsciiString(64);
+	wo->loggingAgentAdminAddress  = "unix:" + wo->generation->getPath() + "/logging_admin";
+
+	UPDATE_TRACE_POINT();
+	wo->adminToolStatusPassword = wo->randomGenerator.generateAsciiString(MESSAGE_SERVER_MAX_PASSWORD_SIZE);
+	wo->adminToolManipulationPassword = wo->randomGenerator.generateAsciiString(MESSAGE_SERVER_MAX_PASSWORD_SIZE);
+	agentsOptions.set("admin_tool_status_password", wo->adminToolStatusPassword);
+	agentsOptions.set("admin_tool_manipulation_password", wo->adminToolManipulationPassword);
 	if (geteuid() == 0 && !userSwitching) {
-		createFile(generation->getPath() + "/passenger-status-password.txt",
-			adminToolStatusPassword, S_IRUSR, defaultUid, defaultGid);
-		createFile(generation->getPath() + "/admin-manipulation-password.txt",
-			adminToolManipulationPassword, S_IRUSR, defaultUid, defaultGid);
+		createFile(wo->generation->getPath() + "/passenger-status-password.txt",
+			wo->adminToolStatusPassword, S_IRUSR, wo->defaultUid, wo->defaultGid);
+		createFile(wo->generation->getPath() + "/admin-manipulation-password.txt",
+			wo->adminToolManipulationPassword, S_IRUSR, wo->defaultUid, wo->defaultGid);
 	} else {
-		createFile(generation->getPath() + "/passenger-status-password.txt",
-			adminToolStatusPassword, S_IRUSR | S_IWUSR);
-		createFile(generation->getPath() + "/admin-manipulation-password.txt",
-			adminToolManipulationPassword, S_IRUSR | S_IWUSR);
+		createFile(wo->generation->getPath() + "/passenger-status-password.txt",
+			wo->adminToolStatusPassword, S_IRUSR | S_IWUSR);
+		createFile(wo->generation->getPath() + "/admin-manipulation-password.txt",
+			wo->adminToolManipulationPassword, S_IRUSR | S_IWUSR);
 	}
 }
 
 static void
-initializeAgentWatchers() {
+initializeAgentWatchers(const WorkingObjectsPtr &wo, vector<AgentWatcherPtr> &watchers) {
 	TRACE_POINT();
-	watchers.push_back(make_shared<HelperAgentWatcher>(*resourceLocator));
-	watchers.push_back(make_shared<LoggingAgentWatcher>(*resourceLocator));
+	watchers.push_back(make_shared<HelperAgentWatcher>(wo));
+	watchers.push_back(make_shared<LoggingAgentWatcher>(wo));
 }
 
 static void
-startAgents() {
+startAgents(const WorkingObjectsPtr &wo, vector<AgentWatcherPtr> &watchers) {
 	TRACE_POINT();
 	foreach (AgentWatcherPtr watcher, watchers) {
 		try {
@@ -566,10 +696,10 @@ startAgents() {
 }
 
 static void
-startWatchingAgents() {
+beginWatchingAgents(const WorkingObjectsPtr &wo, vector<AgentWatcherPtr> &watchers) {
 	foreach (AgentWatcherPtr watcher, watchers) {
 		try {
-			watcher->startWatching();
+			watcher->beginWatching();
 		} catch (const std::exception &e) {
 			writeArrayMessage(FEEDBACK_FD,
 				"Watchdog startup error",
@@ -583,13 +713,13 @@ startWatchingAgents() {
 }
 
 static void
-reportAgentsInformation() {
+reportAgentsInformation(const WorkingObjectsPtr &wo, const vector<AgentWatcherPtr> &watchers) {
 	TRACE_POINT();
 	VariantMap report;
 
 	report
-		.set("server_instance_dir", serverInstanceDir->getPath())
-		.setInt("generation", generation->getNumber());
+		.set("server_instance_dir", wo->serverInstanceDir->getPath())
+		.setInt("generation", wo->generation->getNumber());
 
 	foreach (AgentWatcherPtr watcher, watchers) {
 		watcher->reportAgentsInformation(report);
@@ -602,13 +732,18 @@ int
 main(int argc, char *argv[]) {
 	initializeBareEssentials(argc, argv);
 	P_DEBUG("Starting Watchdog...");
+	WorkingObjectsPtr wo;
+	ServerInstanceDirToucherPtr serverInstanceDirToucher;
+	vector<AgentWatcherPtr> watchers;
 	
 	try {
 		TRACE_POINT();
 		initializeOptions();
 		maybeSetsid();
-		initializeWorkingObjects();
-		initializeAgentWatchers();
+		initializeWorkingObjects(wo, serverInstanceDirToucher);
+		initializeAgentWatchers(wo, watchers);
+		UPDATE_TRACE_POINT();
+		runHookScriptAndThrowOnError("before_watchdog_initialization");
 	} catch (const std::exception &e) {
 		writeArrayMessage(FEEDBACK_FD,
 			"Watchdog startup error",
@@ -620,15 +755,17 @@ main(int argc, char *argv[]) {
 
 	try {
 		TRACE_POINT();
-		startAgents();
-		startWatchingAgents();
-		reportAgentsInformation();
+		startAgents(wo, watchers);
+		beginWatchingAgents(wo, watchers);
+		reportAgentsInformation(wo, watchers);
 		P_INFO("All Phusion Passenger agents started!");
+		UPDATE_TRACE_POINT();
+		runHookScriptAndThrowOnError("after_watchdog_initialization");
 		
 		UPDATE_TRACE_POINT();
 		this_thread::disable_interruption di;
 		this_thread::disable_syscall_interruption dsi;
-		bool exitGracefully = waitForStarterProcessOrWatchers(watchers);
+		bool exitGracefully = waitForStarterProcessOrWatchers(wo, watchers);
 		if (exitGracefully) {
 			/* Fork a child process which cleans up all the agent processes in
 			 * the background and exit this watchdog process so that we don't block
@@ -639,16 +776,19 @@ main(int argc, char *argv[]) {
 			P_DEBUG("Web server did not exit gracefully, forcing shutdown of all agents...");
 		}
 		UPDATE_TRACE_POINT();
+		runHookScriptAndThrowOnError("after_watchdog_shutdown");
+		UPDATE_TRACE_POINT();
 		AgentWatcher::stopWatching(watchers);
 		if (exitGracefully) {
 			UPDATE_TRACE_POINT();
-			cleanupAgentsInBackground(watchers, argv);
-			return 0;
+			cleanupAgentsInBackground(wo, watchers, argv);
 		} else {
 			UPDATE_TRACE_POINT();
 			forceAllAgentsShutdown(watchers);
-			return 1;
 		}
+		UPDATE_TRACE_POINT();
+		runHookScriptAndThrowOnError("after_watchdog_shutdown");
+		return exitGracefully ? 0 : 1;
 	} catch (const tracable_exception &e) {
 		P_ERROR(e.what() << "\n" << e.backtrace());
 		return 1;
